import java.net.*;
import java.util.*;
import java.text.*;
import java.io.*;

/**
 * @author Sergio Blanco Cuaresma
 *
 * Averagin Algorithm: Algorisme descentralitzat de sincrinitzaci de rellotges.
 */
public class Sync {	

	public static void main(String[] args) throws Exception {
		// Parametres de control
		int timeout = 5000;	// miliseconds que escoltem broadcasts
		int broadcastPeriod = 10000; // miliseconds que triguem en tornar a fer broadcast
		int port = 2000;
		String destinationIP = "255.255.255.255"; // Ficar addrea de broadcast de la xarxa
		
		// Calculs per ajustar rellotge
		long deviation = 0;
		long deviationAvg = 0;
		
		// Formatejadors de dades
		SimpleDateFormat formatter = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");	// Broadcast
		SimpleDateFormat formatterCommand = new SimpleDateFormat("MMddHHmmyyyy.ss");	// Comanda sistema
		
		// Socket per fer UDP Broadcast
		DatagramSocket clientSocket = new DatagramSocket(port);
		InetAddress IPAddress = InetAddress.getByName(destinationIP);
		byte[] receiveData = new byte[1024];
		byte[] sendData = new byte[1024];
		
		while (true) {
			long fiPeriode = new Date().getTime() + broadcastPeriod;
			
			// Fem broadcast de l'hora actual
			Date now = new Date();
			String send = formatter.format(now);
			System.out.println("Broadcast: " + send);
			sendData = send.getBytes();
			DatagramPacket sendPacket = new DatagramPacket(sendData, sendData.length, IPAddress, port);		
			clientSocket.send(sendPacket);
			
			// Escoltem broadcasts d'altres durant "timeout"
			System.out.println("Escoltem...");
			
			DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
			try {
				deviationAvg = 0;
				// Haurem de parar d'escoltar dintre de "now + timeout" segons
				long stopSec = new Date().getTime() + timeout;
				clientSocket.setSoTimeout(timeout);
				while (true) {
					// Escoltem
					clientSocket.receive(receivePacket);
			
					// Procesem dades
					String responseTime = new String(receivePacket.getData());
					System.out.println("Rebut: " + responseTime);
					Date pseudoNow = formatter.parse(responseTime);
					// Calculem la desviaci mitja
					deviation = new Date().getTime() - pseudoNow.getTime();
					deviationAvg = (deviationAvg + deviation) / 2;
					
					// Comprovem quant falta per acabar el periode
					long nowSec = new Date().getTime();
					long leftSec = stopSec - nowSec;
					if (leftSec > 0) {
						// El timeout del socket es reseteja al rebre dades, el canviem per
						// que esperi noms el que falta.
						clientSocket.setSoTimeout((int)leftSec);
					} else {
						// S'ha acabat el periode
						break;
					}
				}
			} catch (SocketTimeoutException e) {
				System.out.println("... fi escoltar.");
			}
			
			// Ajustem l'hora del sistema
			Date correctTime = new Date();
			correctTime.setTime(new Date().getTime() + deviationAvg);
			
			System.out.println("Hora ajustada (+" + deviationAvg + "ms): " + formatter.format(correctTime));
			// Executem comanda "date" de linux per modificar l'hora del sistema
			String command[]= {"date", formatterCommand.format(correctTime)};
			exec(command);
			
			// Dormim fins que ens toqui tornar a fer broadcast
			long leftPeriode = fiPeriode - new Date().getTime();
			System.out.println("Dormint... falten " + leftPeriode + "ms de " + broadcastPeriod + "ms");
			if (leftPeriode > 0) {
				Thread.sleep((int)leftPeriode);
			}
		}
	}
	
	// Mtode que ens permet executar comandes del sistema
	static public String[] exec(String[] cmd) throws IOException {
        // Llista per capturar la sortida
        ArrayList list = new ArrayList();

		// Executem la comanda
		Process proc = Runtime.getRuntime().exec(cmd);

		// Accedim al output
		InputStream istr = proc.getInputStream();
		BufferedReader br = new BufferedReader(new InputStreamReader(istr));

		// Llegim la sortida i ho afegim a la llista
		String str;
		while ((str = br.readLine()) != null) {
			list.add(str);
		}

		// Esperem que acabi l'execuci de la comanda
		try {
			proc.waitFor();
		} catch (InterruptedException e) {
			System.err.println("La comanda ha sigut interompuda.");
		}

		// Comprovem si la execuci ha sigut correcta
        if (proc.exitValue() != 0) {
			System.err.println("El valor de sortida no es cero.");
        }

		br.close();

		// Retornem un array d'strings
		return (String[])list.toArray(new String[0]);
	}

}
